package ca.javajeff.lexan;

import java.lang.reflect.Field;

import java.util.ArrayList;
import java.util.List;

import java.util.regex.Matcher;

/**
 *  A lexical analyzer. You can use this class to transform an input stream of 
 *  characters into an output stream of tokens.
 *
 *  @author Jeff Friesen
 */

public final class Lexan
{
   private List<TokLex> tokLexes;

   private Token[] values;

   /**
    *  Initialize a lexical analyzer to a set of Token objects.
    *
    *  @param tokensClass the Class object of a class containing a set of Token
    *         objects
    *
    *  @throws LexanException unable to construct a Lexan object, possibly 
    *          because there are no Token objects in the class
    */

   public Lexan(Class<?> tokensClass) throws LexanException
   {
      try
      {
         tokLexes = new ArrayList<>();
         List<Token> _values = new ArrayList<>();
         Field[] fields = tokensClass.getDeclaredFields();
         for (Field field: fields)
            if (field.getType().getName().equals("ca.javajeff.lexan.Token"))
               _values.add((Token) field.get(null));
         values = _values.toArray(new Token[0]);
         if (values.length == 0)
            throw new LexanException("no tokens");
      }
      catch (IllegalAccessException iae)
      {
         throw new LexanException(iae.getMessage());
      }
   }

   /**
    *  Get this lexical analyzer's list of toklexes.
    *
    *  @return list of toklexes
    */

   public List<TokLex> getTokLexes()
   {
      return tokLexes;
   }

   /**
    *  Lex an input string into a list of toklexes.
    *
    *  @param str the string being lexed
    *
    *  @throws LexException unexpected character found in input
    */

   public void lex(String str) throws LexException
   {
      String s = new String(str).trim(); // remove leading whitespace
      int index = (str.length() - s.length());
      tokLexes.clear();
      while (!s.equals(""))
      {
         boolean match = false;
         for (int i = 0; i < values.length; i++)
         {
            Token token = values[i];
            Matcher m = token.getPattern().matcher(s);
            if (m.find())
            {
               match = true;
               tokLexes.add(new TokLex(token, m.group().trim()));
               String t = s;
               s = m.replaceFirst("").trim(); // remove leading whitespace
               index += (t.length() - s.length());
               break;
            }
         }
         if (!match)
            throw new LexException("Unexpected character in input: " + s, str,
                                   index);
      }
   }
}